/*
 * lookfor.c
 *
 *  Created on: 3 oct. 2012
 *      Author: coissac
 */

#include <string.h>

#include "orgasm.h"
#include "_sse.h"
/**
 * Rigth shift a nucleic sequence encoded with the encode
 * encode.
 *
 * dest : a 16bytes aligned pointer to the place where to store
 *        the shiftedkey. If dest is NULL a static internal buffer
 *        is used.
 *
 * key  : a 16bytes aligned pointer to the key to shift.
 *
 * shift: a value between 0 and 3 indicating the count of base pair
 *        to shift. 0 leads to no action. Value greater than 3 are
 *        reduced to this interval
 *
 * keyLength: a pointer to the length of the compressed key expressed
 *            in bytes. This length is potentially changed for reflecting
 *            the length of the shifted key
 *
 * returns a pointer to the shifted key.
 *
 * WARNING even if dest is not NULL, the returned pointer can be
 * different than dest. So alwaus use the returned pointer. The
 * returned pointer is always 16bytes aligned
 */
pnuc shiftKey(pnuc dest, pnuc key,uint32_t shift, uint32_t keyLength)
{
	static char buffer[512];
	uint32_t  n16bytes;
	uint32_t  remains;
	uint32_t  i;
	um128 out;
	um128 tmp;

	uint8_t mask;
	uint8_t cmask;
	uint8_t rmask;
	uint8_t save;
	uint8_t saveLast;
	uint8_t lshift;
	uint8_t rshift;


	pnuc rep;


	if (dest == NULL)
		dest = (pnuc) PTR16(buffer);

	ASSERT(((size_t)dest & 0xF)==0,"Pointer %p is not 16bytes aligned",dest)

	rep = dest;

	// Shift == 0 is no shift so no computation
	if (shift==0) return key;

	rshift = shift << 1;
	lshift = 8 - rshift;

	mask = (1 << rshift) - 1;
	cmask= ~mask;
	rmask= (~0) >> rshift;

	// The shifted key occupied one extra byte
	keyLength++;

	n16bytes = keyLength >>4;
	remains  = keyLength & 0xF;
	save = 0;

	for (i=0; i < n16bytes; i++, key+=16,dest+=16)
	{
		out.i = _MM_LOAD_SI128((const __m128i*)key);
		saveLast = out.u8[15];
		tmp.i = _MM_SRLI_EPI64(_MM_AND_SI128(out.i,
				 _MM_SET1_EPI8(cmask)),rshift);
		out.i = _MM_SLLI_EPI64(_MM_AND_SI128(out.i,_MM_SET1_EPI8(mask)),
														   lshift);
		out.i = _MM_SLLI_SI128(out.i,1);
		out.i = _MM_OR_SI128(out.i,tmp.i);

		if (i>0)
			out.u8[0]|= save << rshift;
		else
			out.u8[0]&= rmask;

		save = saveLast;

		_MM_STORE_SI128((__m128i*)dest,out.i);
	}

	for (i=0; i < remains; i++, key++,dest++)
	{
		saveLast = *key;
		*dest = (saveLast >> rshift) | (save << lshift);
		save = saveLast;
	}

	return rep;
}

/**
 * Compare the prefix of a key encoded by the encode function over
 * the length 'length' expressed in base pair with a prefix of same
 * length corresponding to the read r1 from buffer.
 *
 * buffer : a pointer to a read buffer
 * r1     : the identifer of the read to compare
 * key    : a pointer to the encoded second sequence involved in the comparison
 * length : the length of the prefix to compare
 *
 * return 0 if both the prefix are equal, -1 if the r1 sequence is lexicographically
 * before key and +1 otherwise.
 *
 */
int8_t cmpPrefix(buffer_t *buffer,uint32_t r1,pnuc key,uint32_t length)
{
	uint32_t shift;
	uint32_t i;

	uint8_t  mask;
	uint8_t  *pr1;
	uint8_t  *pr2;
	uint8_t  vr1=0;
	uint8_t  vr2=0;

	shift = CODELENGTH(length);
	mask  = ~((1 << ((4 - (length & 3)) * 2))-1);

	// DEBUG("pos : %d mask : %d,%x",pos,shift,mask);

	pr1 = ((uint8_t*)(buffer->records)) + buffer->recordSize * r1;
	pr2 = (uint8_t*)key;
	i=0;

	do {
		vr1 = *pr1;
		vr2 = *pr2;

		if (i==(shift-1))
		{
			vr1&=mask;
			vr2&=mask;
		}

		i++; pr1++; pr2++;

	} while( i < shift && vr2==vr1);

	return (vr1==vr2) ? 0 : ((vr1 < vr2) ? -1:1);

}


/**
 * Compare the prefix of a key encoded by the encode function over
 * the length 'length' expressed in base pair with a prefix of same
 * length corresponding to the read r1 from buffer.
 *
 * buffer : a pointer to a read buffer
 * r1     : the identifer of the read to compare
 * key    : a pointer to the encoded second sequence involved in the comparison
 * length : the length of the prefix to compare
 *
 * return 0 if both the prefix are equal, -1 if the r1 sequence is lexicographically
 * before key and +1 otherwise.
 *
 * INTERNAL FUNCTION DO NOT USE DIRECTLY
 *
 */

int8_t cmpCompPrefix(buffer_t *buffer,uint32_t r1,pnuc key,uint32_t lkey, uint32_t length)
{
	int32_t i;
	uint32_t querylength;
	uint32_t readlength;
	uint8_t  mask;
	uint8_t  *pr1;
	uint8_t  *pr2;
	uint8_t  vr1=0;
	uint8_t  vr2=0;

	querylength = CODELENGTH(length);
	readlength  = CODELENGTH(buffer->readSize);
	mask= ((uint8_t)(~0)) >> ((((4-(length & 3))) & 3) << 1);

	// DEBUG("pos : %d mask : %d,%x",pos,shift,mask);

	pr1 = ((uint8_t*)(buffer->records)) + buffer->recordSize * r1 + readlength - 1;
	pr2 = ((uint8_t*)(key)) + lkey - 1;

	i=querylength-1;

	do {
		vr1 = *pr1;
		vr2 = *pr2;

		if (i==0)
		{
			vr1&=mask;
			vr2&=mask;
		}

		i--; pr1--; pr2--;

	} while( i >=0 && vr2==vr1);

	vr1 = complement4nuc[vr1];
	vr2 = complement4nuc[vr2];

	return (vr1==vr2) ? 0 : ((vr1 < vr2) ? -1:1);

}


int32_t lookForForward(buffer_t *buffer, pnuc key, size_t length, int32_t* endoflist)
{
	int32_t initcode;
	int32_t start;
	int32_t end;
	int32_t middle;
	int32_t comp;
	int32_t comp1;

	// I set the end of list flag to 0
	*endoflist=0;

	// and extract the 8 first nucleotides
	initcode = *((uint16_t*)key);

#ifdef LITTLE_END
	initcode = ((initcode & 255) << 8) | (initcode >> 8);
#endif

	// I use the hash table for identifying positions of reads
	// starting with these 8 nucleotides

	start=buffer->index1[initcode];

	if (initcode < 0xFFFF)
		end = buffer->index1[initcode+1];
	else
		end = buffer->readCount;


	// in this case no reads begin with the 8mer
	if (end-start==0)
	{
		*endoflist=1;
		return 0;
	}

	// I start a bsearch between start and end

//	middle = (start + end) / 2;
//	comp = cmpPrefix(buffer,middle,key,length);

	end--;

	while (start!=end)
	{
		middle = (start + end) / 2;
		comp = cmpPrefix(buffer,middle,key,length);

		if (comp < 0) start=middle+1;
		if (comp >= 0) end=middle;

	}

	comp = cmpPrefix(buffer,start,key,length);

	// No match found

	if (comp!=0)
	{
		*endoflist=1;
		return 0;
	}

	return start;

}

int32_t nextForward(buffer_t *buffer, int32_t current, size_t length, int32_t* endoflist)
{

	int32_t pos= current+1;
	pnuc    key= (pnuc)((uint8_t*)(buffer->records) + buffer->recordSize * current);
	int     comp=cmpPrefix(buffer,pos,key,length);

	if (comp==0)
		return pos;

	*endoflist=1;
	return 0;
}

int32_t lookForReverse(buffer_t *buffer, pnuc key, uint32_t lkey, size_t length, int32_t* endoflist)
{
	char internalbuffer[512];
	char internalbuffer2[512];
	pnuc dest;
	pnuc dest2;
	int32_t shift;
	int32_t rshift;
	uint32_t i;
	int32_t initcode;
	int32_t start;
	int32_t end;
	int32_t middle;
	int32_t comp;

	// I set the end of list flag to 0
	*endoflist=0;

	dest = (pnuc) PTR16(internalbuffer);
	dest2 = (pnuc) PTR16(internalbuffer2);
	shift= 4 - (length & 3);

	// then we reverse complement the sequence

	for (i=0; i<lkey;i++)
		dest2[i]=complement4nuc[(uint8_t)key[lkey-1-i]];

	// We have now to "bit align" the end of the key with the end of the reads

	rshift= buffer->readSize & 3;

	if (rshift > 0)
	{
		dest = shiftKey(dest, dest2, rshift, lkey);
		lkey++;

		// we have to increase match length to take into account the
		// unfully filled last byte of the reads and key

		length+= 4 - rshift;
	}
	else
		dest = dest2;

	// Now dest point to the key to use

	initcode = *((uint16_t*)(dest + lkey)-1);
#ifdef LITTLE_END
	initcode = complement4nuc[initcode & 0x00FF] | (complement4nuc[initcode >> 8] << 8);
#else
	initcode = complement4nuc[(initcode & 255)] << 8 | complement4nuc[initcode >> 8];
#endif

	// I use the hash table for identifying positions of reads
	// starting with these 8 nucleotides

	start=buffer->index2[initcode];

	if (initcode < 0xFFFF)
		end = buffer->index2[initcode+1];
	else
		end = buffer->readCount;


	// in this case no reads begin with the 8mer
	if (end-start==0)
	{
		*endoflist=1;
		return 0;
	}

	// I start a bsearch between start and end

	end--;

	while (start!=end)
	{
		middle = (start + end) / 2;
		comp = cmpCompPrefix(buffer,buffer->order1[middle],dest,lkey,length);

		if (comp < 0) start=middle+1;
		if (comp >= 0) end=middle;

	}

	comp = cmpCompPrefix(buffer,buffer->order1[start],dest,lkey,length);

	// No match found

	if (comp!=0)
	{
		*endoflist=1;
		return 0;
	}


	return start;

}

int32_t nextReverse(buffer_t *buffer, int32_t current, size_t length, int32_t* endoflist)
{
	int32_t pos= current+1;
	int32_t lkey = CODELENGTH(buffer->readSize);
	pnuc    key= (pnuc)((uint8_t*)(buffer->records) + buffer->recordSize * buffer->order1[current]);
	int32_t rshift= buffer->readSize & 3;
	int     comp=cmpCompPrefix(buffer,buffer->order1[pos],key,lkey, length + ((rshift==0) ? 0:(4-rshift)));

	if (comp==0)
		return pos;

	*endoflist=1;
	return 0;

}

int32_t lookForReads(buffer_t *buffer, pnuc key, int32_t lkey,size_t length, int32_t* endoflist)
{
	int32_t pos = lookForForward(buffer, key, length, endoflist);

	if (*endoflist==0)
		pos++;
	else
	{
		pos = lookForReverse(buffer, key, lkey, length, endoflist);
		if (*endoflist==0)
			pos = -pos - 1;
	}

	return pos;
}

int32_t nextRead(buffer_t *buffer, int32_t current, size_t length, int32_t* endoflist)
{
	int32_t next;
	pnuc from;

	if (current > 0)
	{
		current--;
		next = nextForward(buffer, current, length, endoflist);

		if (*endoflist==0)
			next++;
		else
		{
			from = (pnuc)((uint8_t*)(buffer->records) + buffer->recordSize * current);

			next = lookForReverse(buffer, from, CODELENGTH(buffer->readSize),length, endoflist);

			if (*endoflist==0)
				next = -next -1;
		}

	}

	if (current < 0)
	{
		current = -current -1;
		next = nextReverse(buffer, current, length, endoflist);

		if (*endoflist==0)
			next = -next -1;
	}

	if (current == 0)
		*endoflist=0;

	return next;
}

int32_t lookForString(buffer_t *buffer, char *key, size_t length, int32_t* endoflist)
{
	char internalbuffer[512];
	pnuc dest;

	ASSERT(length>=8,"length have to be greater or equal to %d",8)


	// I encode the key string
	dest = (pnuc) PTR16(internalbuffer);
	dest = encodeSequence(dest, key, length);

	return lookForReads(buffer, dest, CODELENGTH(length), length, endoflist);
}
